""" file:   singleton.py(pysiss.utilities)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   25 August 2014

    description: A Singleton class instance, add __metaclass__ = Singleton to
    your class to use it
"""


class Singleton(type):

    """ A singleton metaclass for implementing registries.

        This metaclass implements the Singleton pattern, so that only one
        instance of a class is ever instantiated. Subsequent calls to
        `__init__` will return a reference to this instantiation. To use
        this in your classes, just add

            __metaclass__ = Singleton

        to your class definition.
    """

    def __init__(cls, name, bases, dictionary):
        super(Singleton, cls).__init__(name, bases, dictionary)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance



""" file:   unmarshallers.py (pysiss.vocabulary.gml)
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: Unmarshalling functions for GeoSciML/GML objects
"""

from shapely.geometry import Polygon, LineString
from ..namespaces import NamespaceRegistry

NAMESPACES = NamespaceRegistry()


def position(elem):
    """ Unmarshal a gml:posList, gml:pos or gml:coordinates element
    """
    if elem.text:
        token_pairs = elem.text.split('\n')[1:-1]
        return [map(float, p.split()) for p in token_pairs]
    else:
        return None


def polygon(elem):
    """ Unmarshal a gml:Polygon element
    """
    # Get the projection
    projection = elem.xpath('.//@srsName',
                            namespaces=NAMESPACES)[0]

    # Get outer boundary first, we always have this
    outer = position(
        elem.xpath('.//gml:outerBoundaryIs//gml:posList',
                   namespaces=NAMESPACES)[0])

    # We may have 0, 1 or more inner boundaries
    inners = elem.xpath('.//gml:innerBoundaryIs//gml:posList',
                        namespaces=NAMESPACES)
    if not inners:
        inners = None
    else:
        inners = map(position, inners)

    return {'projection': projection,
            'shape': Polygon(outer, inners)}


def linestring(elem):
    """ Unmarshal a gml:LineString element
    """
    # Get the projection
    projection = elem.xpath('.//@srsName',
                            namespaces=NAMESPACES)[0]

    # Get the LineString text
    string = position(
        elem.xpath('./gml:posList',
                   namespaces=NAMESPACES)[0])

    return {'projection': projection,
            'shape': LineString(string)}


def description(elem):
    """ Unmarshal a gml:description element
    """
    return elem.text


UNMARSHALLERS = {
    'gml:posList': position,
    'gml:pos': position,
    'gml:coordinates': position,
    'gml:Polygon': polygon,
    'gml:LineString': linestring,
    'gml:description': description,
}

__all__ = (position, polygon, linestring, UNMARSHALLERS)



""" file: property.py (pysiss.borehole.properties)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Imports for pysiss.borehole.properties
"""

class Property(object):

    """ Container for values with type.

        Values must match the length of the dataset: for sampling and interval
        datasets, it must be a sequence of the same length as the depths. For a
        feature is should be a single value unless it is a multivalued category

        :param property_type: The property metadata for the property
        :type property_type: pysiss.borehole.properties.property_type
        :param values: A list of values to store
        :type values: iterable
    """

    def __init__(self, property_type, values):
        self.property_type = property_type
        self.values = values

    def __repr__(self):
        info = 'Property {0}: {1} values in units of {2}'
        return info.format(self.name, len(self.values),
            self.property_type.units)

    @property
    def name(self):
        """ Returns the name of the property"
        """
        return self.property_type.name

    def copy(self):
        """ Return a copy of the Property instance
        """
        return Property(self.property_type, self.values[:])



""" file:   __init__.py (pysiss.borehole tests)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Initialisation of tests.
"""

import pkgutil
import inspect

__all__ = []

# Discover all test files in test directory, and inject their
# global namespaces into the current global namespace. This
# lets unittest discover the tests nicely
for loader, name, is_pkg in pkgutil.walk_packages(__path__):
    module = loader.find_module(name).load_module(name)

    for name, value in inspect.getmembers(module):
        if name.startswith('__'):
            continue

        globals()[name] = value
        __all__.append(name)



""" file:   collection.py (pysiss.utilities)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   25 August 2014

    description: A utility class for forming collections of objects
"""


class Collection(list):

    """ A collection of objects, accessible as a list or dictionary

        :param objects: The objects to add on initialization
        :type objects: list of object instances
    """

    def __init__(self, objects=None):
        super(Collection, self).__init__()
        self._index = {}

        # Add the list of objects if required
        if objects:
            for bh in objects:
                self.append(bh)

    def __getitem__(self, ident_or_idx):
        """ Retrieve a object from the collection

            :param ident_or_idx: Either an integer index, or a object name.
        """
        try:
            return super(Collection, self).__getitem__(ident_or_idx)
        except IndexError:
            try:
                return super(Collection, self).__getitem__(self._index[ident_or_idx])
            except KeyError:
                str = ('Unknown key or index {0} passed '
                       'to BoreholeCollection').format(ident_or_idx)
                raise IndexError(str)

    def __setitem__(self, object):
        """ Add a object to the collection
        """
        self.append(object)
        self._index[object.name] = len(self)

    def __delitem__(self, object_name):
        """ Remove a object from the collection
        """
        # Find location of object in list
        idx = self._index[object_name]

        # Move subsequent object indices up
        for obj in self[idx:]:
            self._index[obj.name] -= 1

        # Delete object
        del self._index[object_name]
        del self[idx]

    def keys(self):
        return [obj.name for obj in self]

    def values(self):
        return self

    def items(self):
        return zip(self.keys(), self.values())

    @property
    def shapes(self):
        return (obj.shape for obj in self)



""" file:   __init__.py (pysiss.vocabulary.erml)
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: init file
"""

import unmarshallers

__all__ = [unmarshallers]



""" file: objects.py (pysiss.utilities)

    description: Some basic metaclasses etc for defining pysiss classes
"""

import uuid


class id_object(object):

    """ A mixin class to implement UUID comparisons for child classes

        This metaclass generates a UUID for a class at initialization,
        and defines the class __eq__ method to use this UUID.
    """

    def __init__(self, name, *args, **kwargs):
        super(id_object, self).__init__(*args, **kwargs)
        self.uuid = uuid.uuid5(uuid.NAMESPACE_DNS, name)

    def __eq__(self, other):
        """ Equality test

            Class instances are equal if their UUIDs match
        """
        return self.uuid == other.uuid



""" file: maths.py (pysiss.utilities)
"""

import numpy
import functools


def heaviside(values):
    r""" Heaviside function.

        The Heaviside step function, or the unit step function, usually
        denoted by :math`H`, is a discontinuous function whose value is zero
        for negative argument and one for positive argument. We also take
        :math:`H(0) = 0`, although this rarely matters in practise.

        :param values: the argument values
        :type values: numpy.ndarray
        :return: the Heaviside function evaluated at each value
    """
    result = numpy.ones_like(values)
    result[values <= 0] = 0.
    return result


def same_sign(value1, value2):
    """ Determine whether two values have the same sign

    """
    return numpy.sign(value1, dtype=int) == numpy.sign(value2, dtype=int)


def integrate(times, values):
    """ Return the definite integral of the given data at the given
        times
    """
    times = numpy.asarray(times)
    values = numpy.asarray(values)
    return numpy.cumsum(numpy.gradient(times) * values)


def mask_all_nans(*arrays):
    """ Mask all indices where any array has a NaN.

        Example usage:

            mask = mask_all_nans(array1, array2)
            array1[mask]            # Both of these are guarenteed
            array2[mask]            # to be nan-free

        :param *arrays: The arrays to generate masks for. They should all have
            the same size or a ValueError will be raised.
        :type *arrays: `numpy.ndarrays`
        :returns: A `numpy.boolean_array` which can be used as a mask.
    """
    # Convert arrays to numpy arrays if required
    try:
        arrays = [numpy.asarray(a, dtype=numpy.float_) for a in arrays]
    except ValueError:
        raise ValueError("Arrays supplied to mask_all_nans must be able to be "
                         "converted to floats!")

    # Check that everything has the same shape
    shapes = [a.shape for a in arrays]
    if any([s != shapes[0] for s in shapes]):
        raise ValueError("Arrays supplied to mask_all_nans must be the same "
                         "size (arrays have shapes {0})".format(shapes))

    # Return the non-nan indices
    return numpy.logical_not(functools.reduce(numpy.logical_or,
                             [numpy.isnan(a) for a in arrays]))


def try_float(value_str):
    """ Tries to make str a value, returns NaN if not
    """
    try:
        return float(value_str)
    except ValueError:
        return numpy.nan



#!/usr/bin/env python
""" file:   test_detrend.py
    author: Jess Robertson,
            CSIRO Earth Science and Resource Engineering
    date:   Tuesday November 19, 2013

    description: Tests for detrending functions
"""

import unittest
import numpy
from pysiss.borehole.analysis import detrend


class TestDetrend(unittest.TestCase):

    """ Unit tests for pysiss.borehole.analysis.detrend
    """

    def setUp(self):
        # Data and expected results
        self.data = numpy.linspace(0, 1)
        zeros = numpy.zeros_like(self.data, dtype=numpy.float_)
        self.expected = {
            'none': numpy.linspace(0, 1),
            'mean': numpy.linspace(-0.5, 0.5),
            'linear': zeros,
            'quadratic': zeros,
            'cubic': zeros,
        }

        # Checking for equality between float arrays
        tiny = 1e-10  # should really use numpy.MachAr().tiny
        self.narray_eq = lambda a, b: numpy.all((a - b) ** 2 < tiny)

    def test_detrend_modification(self):
        """ Data should be modified in place
        """
        data = numpy.linspace(0, 1)
        self.assertEqual(detrend(data), None)

    def test_detrend_values(self):
        """ Test values returned by detrend
        """
        for trend, exp_values in self.expected.items():
            data = numpy.linspace(0, 1)  # Need to recreate data after detrend
            self.assertEqual(detrend(data, trend), None)
            self.assertTrue(self.narray_eq(data, exp_values))

    def test_default_detrend(self):
        """ Detrend should default to linear detrending
        """
        data = numpy.linspace(0, 1)
        detrend(data)
        self.assertTrue(self.narray_eq(data, self.expected['linear']))



""" file:   metadata.py (pysiss.metadata)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Wednesday 27 August, 2014

    description: Functions to deal with gsml:geologicFeature data

    Geologic features can be shared by many different objects, so it makes
    sense to seperate these out into a seperate registry.
"""

from ..utilities import id_object
from .registry import MetadataRegistry


class Metadata(id_object):

    """ Class to store metadata record
    """

    registry = MetadataRegistry()

    def __init__(self, ident, tree, type, **kwargs):
        super(Metadata, self).__init__(name=type)
        self.ident = ident or self.uuid
        self.tree = tree
        self.type = type

        # Store other metadata
        for attrib, value in kwargs.items():
            setattr(self, attrib, value)

        # Register yourself with the registry
        self.registry.register(self)

    def __str__(self):
        return 'Metadata record {0}, of type {1}'.format(self.ident, self.type)

    def xpath(self, *args, **kwargs):
        """ Pass XPath queries through to underlying tree
        """
        return self.tree.xpath(*args, **kwargs)



""" file:   point_dataset.py (pysiss.borehole.datasets)
    author: Jess Robertson & Ben Caradoc-Davies
            CSIRO Earth Science and Resource Engineering
    date:   Sunday November 10, 2013

    description: DataSet for point sample data (data sampled at a finite set of
            points down a borehole)
"""

from .dataset import DataSet

import numpy
from scipy.interpolate import InterpolatedUnivariateSpline as Spline
import pandas


class PointDataSet(DataSet):

    """ DataSet for data defined at points in the dataset.

        A PointDataSet is a sequence of depths at which continuous
        properties are sampled. Analogous to a coverage. One PointDataSet
        can be interpolated onto another.

        Depths must be in monotonically increasing order.

        :param name: identifier for the dataSet
        :type name: string
        :param depths: sample down-hole depths in metres from collar
        :type depths: iterable
        :param details: The metadata associated with the dataset. Optional,
            defaults to None.
        :type details: pysiss.borehole.dataset.DatasetDetails
    """

    def __init__(self, name, depths, details=None):
        super(PointDataSet, self).__init__(
            name, len(depths), details=details)
        depths = numpy.asarray(depths)
        assert all(numpy.gradient(depths) > 0), \
            "depths must be monotonically increasing"
        self.depths = depths

    def __repr__(self):
        info = 'PointDataSet {0}: with {1} depths and {2} '\
               'properties'
        return info.format(self.name, len(self.depths),
                           len(self.properties))

    def get_interval(self, from_depth, to_depth, dataset_name=None):
        """ Return the data between the given depths as as new PointDataSet
        """
        # Specify a name if not already passed
        if dataset_name is None:
            dataset_name = '{0}: subdataset {1} to {2}'.format(
                self.name, from_depth, to_depth)

        # Generate a new PointDataSet
        indices = self.get_interval_indices(from_depth, to_depth)
        newdom = PointDataSet(dataset_name, self.depths[indices])
        for prop in self.properties.values():
            newdom.add_property(prop.property_type, prop.values[indices])
        return newdom

    def get_interval_indices(self, from_depth, to_depth):
        """ Returns the indices for the depths in the given interval
        """
        return numpy.where(
            numpy.logical_and(self.depths >= from_depth,
                              self.depths <= to_depth))[0]

    def split_at_gaps(self, gap_metric='spacing_median', threshold=10):
        """ Split a dataset by finding significant gaps in the dataset.

            A metric to define 'gaps' is required. Currently we only have
            'spacing_median', which is really only suitable for PointDataSet
            instances. A gap is found where the sample spacing is greater than
            threshold * <gap_metric>.

            Available metrics:
                'spacing_median': a gap is defined as a spacing between
                    samples which is an order of magnitude above the median
                    sample spacing in a dataset.
        """
        # We need to add a small amount to the dataset so that the interval
        # picker works well in the case of subdatasets with only one value
        epsilon = 1e-10
        jitter = lambda a, b: (a - epsilon, b + epsilon)

        # Select gap metric to use, generate gap locations
        depths = self.depths
        spacing = numpy.diff(depths)
        if gap_metric is 'spacing_median':
            med_spacing = numpy.median(spacing)
            gap_indices = numpy.flatnonzero(spacing > threshold * med_spacing)
        else:
            raise NotImplementedError(
                "Unknown gap metric {0}".format(gap_metric))

        # Aggregate gap intervals
        self.gaps = []
        for idx in gap_indices:
            self.gaps.append((depths[idx], depths[idx + 1]))

        # We need to include the start and end of the dataset!
        gap_indices = [0] + list(gap_indices) + [-1]

        # Form a list of data subdatasets
        self.subdatasets = []
        for idx in range(len(gap_indices) - 1):
            # DataSets start _after_ the gap & end with next gap
            from_depth = depths[gap_indices[idx] + 1]
            to_depth = depths[gap_indices[idx + 1]]

            # Check whether we need to jitter
            if gap_indices[idx] + 1 == gap_indices[idx + 1]:
                from_depth, to_depth = jitter(from_depth, to_depth)
            self.subdatasets.append((from_depth, to_depth))
        return self.subdatasets, self.gaps

    def regularize(self, npoints=None, dataset_name=None, fill_method='median',
                   degree=0):
        """ Resample dataset onto regular grid.

            This regularizes the data so that all sample spacings are equal to
            the median spacing of the raw data. Returns a new PointDataSet
            instance with the new data.

            Arguments:
                npoints - the number of new points. Optional, if not specified
                    then the reinterpolated data will have the median spacing
                    of the raw data.
                dataset_name - the name for the returned PointDataSet.
                    Optional, defaults to "<current_name> resampled".
                fill_method - the method for filling the gaps. 'interpolate'
                    uses the interpolated spline, 'mean' fills gaps with the
                    mean value for the borehole.
                degree - the degree of the interpolation. Optional, defaults
                    to 1 (i.e. linear interpolation). Values > 0 denote
                    polynomial interpolation, a value of 0 uses nearest-
                    neighbour interpolation.
        """
        # We need to identify gaps first
        if self.gaps is None:
            print ("Warning - your dataset hasn't been analysed for gaps yet. "
                   "I'm going to assume you just want to use the default "
                   "values")
            self.split_at_gaps()

       # Specify name & number of points if not already passed
        if dataset_name is None:
            dataset_name = '{0} resampled'.format(self.name)
        if npoints is None:
            spacing = float(numpy.median(numpy.diff(self.depths)))
            npoints = abs(self.depths[-1] - self.depths[0]) / spacing

        # Generate a new DataSet with the resampled data
        new_depths = numpy.linspace(self.depths[0], self.depths[-1], npoints)
        newdom = PointDataSet(dataset_name, new_depths)

        # If we're doing nearest neighbours then we only need to work out the
        # interpolation once
        if degree == 0:
            # This line generates a set of indices which will reconstruct a
            # new signal using nearest neighbours, just do:
            # property.values[interp_indices]
            interp_indices = numpy.argmin(
                numpy.asarray([
                    (self.depths - new_depths[:, numpy.newaxis]) ** 2]),
                axis=-1)[0]

        # Get gap indices etc and store for faster lookup
        if fill_method in ['mean', 'median', 'local mean', 'local median']:
            # These methods need gap indices
            gap_idxs = [newdom.get_interval_indices(*gap) for gap in self.gaps]
        if fill_method in ['local mean', 'local median']:
            # These methods need subdataset indices from the old dataset
            sdom_idxs = [self.get_interval_indices(*sdom)
                         for sdom in self.subdatasets]

        # Resample properties
        for prop in self.properties.values():
            if prop.property_type.isnumeric is False:
                # We can't interpolate non-numeric data
                print ("Property {0} in dataset {1} is not numeric so I'm "
                       "skipping it. If this is a suprise to you, maybe you "
                       "should check whether you've correctly set the "
                       "is_numeric flag in the PropertyType class for this "
                       "property."
                       ).format(prop.property_type.name, self.name)
                continue

            # Generate spline fit if required, else use nearest-neighbours
            if degree == 0:
                new_values = prop.values[interp_indices]
            else:
                spl = Spline(self.depths, prop.values, k=degree)
                new_values = spl(new_depths)

            # Deal with gaps
            if fill_method == 'interpolate':
                # We've already generated an interpolated value, so move on
                # This option is here for error checking purposes
                pass

            elif fill_method == 'mean':
                # Mean value in gaps, poly interp otherwise
                meanval = prop.values.mean()
                for gidx in gap_idxs:
                    new_values[gidx] = meanval

            elif fill_method == 'median':
                # Median value in gaps
                medval = numpy.median(prop.values)
                for gidx in gap_idxs:
                    new_values[gidx] = medval

            elif fill_method == 'local mean':
                # local mean value in gaps
                smeans = [prop.values[s].mean() for s in sdom_idxs]
                for sma, gidx, smb in zip(smeans[:-1], gap_idxs, smeans[1:]):
                    new_values[gidx] = (sma + smb) / 2.

            elif fill_method == 'local median':
                # local median value in gaps
                gap_neighbours = zip(sdom_idxs[:-1], gap_idxs, sdom_idxs[1:])
                for sidxa, gidx, sidxb in gap_neighbours:
                    new_values[gidx] = numpy.median(numpy.concatenate(
                        prop.values[sidxa],
                        prop.values[sidxb]))

            else:
                raise NotImplementedError

            # Push back to new dataset
            newdom.add_property(prop.property_type, new_values)

        # Copy over gaps and subdatasets
        newdom.gaps = self.gaps
        newdom.subdatasets = self.subdatasets
        return newdom

    def resample(self, new_depths, dataset_name=None, fill_method='median',
                 degree=0):
        """ Resample dataset onto regular grid.

            Returns a new PointDataSet instance with the new data.

            Arguments:
                new_depths - The new depths to resample at.
                dataset_name - the name for the returned PointDataSet.
                    Optional, defaults to "<current_name> resampled".
                fill_method - the method for filling the gaps. 'interpolate'
                    uses the interpolated spline, 'mean' fills gaps with the
                    mean value for the borehole.
                degree - the degree of the interpolation. Optional, defaults
                    to 1 (i.e. linear interpolation). Values > 0 denote
                    polynomial interpolation, a value of 0 uses nearest-
                    neighbour interpolation.
        """
        # We need to identify gaps first
        if self.gaps is None:
            print ("Warning - your dataset hasn't been analysed for gaps yet. "
                   "I'm going to assume you just want to use the default "
                   "values")
            self.split_at_gaps()

       # Specify name & number of points if not already passed
        if dataset_name is None:
            dataset_name = '{0} resampled'.format(self.name)

        # Generate a new DataSet with the resampled data
        newdom = PointDataSet(dataset_name, new_depths)

        # If we're doing nearest neighbours then we only need to work out the
        # interpolation once
        if degree == 0:
            # This line generates a set of indices which will reconstruct a
            # new signal using nearest neighbours, just do:
            # property.values[interp_indices]
            interp_indices = numpy.argmin(
                numpy.asarray([
                    (self.depths - new_depths[:, numpy.newaxis]) ** 2]),
                axis=-1)[0]

        # Get gap indices etc and store for faster lookup
        if fill_method in ['mean', 'median', 'local mean', 'local median']:
            # These methods need gap indices
            gap_idxs = [newdom.get_interval_indices(*gap) for gap in self.gaps]
        if fill_method in ['local mean', 'local median']:
            # These methods need subdataset indices from the old dataset
            sdom_idxs = [self.get_interval_indices(*sdom)
                         for sdom in self.subdatasets]

        # Resample properties
        for prop in self.properties.values():
            if prop.property_type.isnumeric is False:
                # We can't interpolate non-numeric data
                print ("Property {0} in dataset {1} is not numeric so I'm "
                       "skipping it. If this is a suprise to you, maybe you "
                       "should check whether you've correctly set the "
                       "is_numeric flag in the PropertyType class for this "
                       "property."
                       ).format(prop.property_type.name, self.name)
                continue

            # Generate spline fit if required, else use nearest-neighbours
            if degree == 0:
                new_values = prop.values[interp_indices]
            else:
                spl = Spline(self.depths, prop.values, k=degree)
                new_values = spl(new_depths)

            # Deal with gaps
            if fill_method == 'interpolate':
                # We've already generated an interpolated value, so move on
                # This option is here for error checking purposes
                pass

            elif fill_method == 'mean':
                # Mean value in gaps, poly interp otherwise
                meanval = prop.values.mean()
                for gidx in gap_idxs:
                    new_values[gidx] = meanval

            elif fill_method == 'median':
                # Median value in gaps
                medval = numpy.median(prop.values)
                for gidx in gap_idxs:
                    new_values[gidx] = medval

            elif fill_method == 'local mean':
                # local mean value in gaps
                smeans = [prop.values[s].mean() for s in sdom_idxs]
                for sma, gidx, smb in zip(smeans[:-1], gap_idxs, smeans[1:]):
                    new_values[gidx] = (sma + smb) / 2.

            elif fill_method == 'local median':
                # local median value in gaps
                gap_neighbours = zip(sdom_idxs[:-1], gap_idxs, sdom_idxs[1:])
                for sidxa, gidx, sidxb in gap_neighbours:
                    new_values[gidx] = numpy.median(numpy.concatenate(
                        prop.values[sidxa],
                        prop.values[sidxb]))

            else:
                raise NotImplementedError

            # Push back to new dataset
            newdom.add_property(prop.property_type, new_values)

        # Copy over gaps and subdatasets
        newdom.gaps = self.gaps
        newdom.subdatasets = self.subdatasets
        return newdom

    def to_dataframe(self):
        """ Tranform the data in the dataset into a Pandas dataframe.
        """
        return pandas.DataFrame(
            data=dict(((k, self.properties[k].values)
                       for k in self.properties.keys())),
            index=self.depths)



""" file:   borehole.py (pysiss.borehole)
    author: Jess Robertson & Ben Caradoc-Davies
            CSIRO Earth Science and Resource Engineering
    date:   September 10, 2013

    description: Borehole class implementation
"""

from .details import Details, detail_type
from .datasets import DataSet, PointDataSet, IntervalDataSet
from .properties import Property
from ..utilities import id_object


class Borehole(id_object):

    """ Class to represent a borehole.

        Borehole has point features and datasets on which properties are
        defined. A property can be defined on multiple datasets. Features and
        datasets are containers for the properties defined on them.

        A Feature is analogous to a spatial point feature. It has a depth and
        properties but it makes no sense to perform any interpolation on these.

        An IntervalDataSet is is a sequence of borehole segments each having a
        single value for each property; this value is taken to be the same
        across the entire length of the interval. IntervalDataSets can be
        merged to form a new IntervalDataSet that has the intervals whose
        boundaries are the union of the boundaries of the source
        IntervalDataSets. An IntervalDataSet can be interpolated onto a
        PointDataSet.

        A PointDataSet is a sequence of depths at which continuous properties
        are sampled. Analogous to a coverage. One PointDataSet can be
        interpolated onto another.

        Depths are measured in metres down-hole from the borehole collar; depth
        sequences must be in monotonically increasing order.

        Property units are expressed as Unified Code for Units of Measure
        (UCUM): http://unitsofmeasure.org/ucum.html

        Some useful properties include:
            features - dict mapping feature name to Feature
            interval_datasets - dict mapping interval dataset name to
                IntervalDataSet
            point_datasets - dict mapping sampling dataset name to
                PointDataSet

        :param name: An identifier for the borehole
        :type name: `string`
        :param origin_position: The borehole's position (lat/long; defaults to
            None)
        :type origin_position: OriginPosition class

    """

    # Mapping dataset types to class attributes
    _type_to_attr = {
        DataSet: 'datasets',
        PointDataSet: 'point_datasets',
        IntervalDataSet: 'interval_datasets',
    }

    def __init__(self, name, origin_position=None):
        super(Borehole, self).__init__(name=name)
        self.name = name
        self.origin_position = origin_position
        self.survey = None
        self.features = dict()
        self.details = BoreholeDetails()

        # Initialize dataset lists
        for dataset_attr in self._type_to_attr.values():
            setattr(self, dataset_attr, dict())

    def __repr__(self):
        """ String representation
        """
        info = 'Borehole {0} at origin position {1} contains '
        info_str = info.format(self.name, self.origin_position)
        n_datasets = sum([len(getattr(self, a))
                         for a in self._type_to_attr.values()])
        summary_str = '{0} datasets'.format(n_datasets)
        summary_str += ' & {0} features'.format(len(self.features))
        dataset_list = ''
        if len(self.interval_datasets) > 0:
            dataset_list += ('\nIDs: ' + '\n     '.join(
                             map(str, self.interval_datasets.values())))
        if len(self.point_datasets) > 0:
            dataset_list += ('\nSDs: ' + '\n     '.join(
                             map(str, self.point_datasets.values())))
        if len(self.details.values()) > 0:
            borehole_details_str = \
                '\nBorehole details: {0}'.format(self.details)
        else:
            borehole_details_str = ''

        return info_str + summary_str + dataset_list + borehole_details_str

    def add_feature(self, name, depth):
        """ Add and return a new Feature.

            :param name: The identifier for the new feature
            :type name: `string`
            :param depth: Down-hole depth in metres from collar
            :type depth: `int` or `float`
            :returns: the new `pysiss.borehole.Feature` instance
        """
        self.features[name] = Feature(name, depth)
        return self.features[name]

    def add_dataset(self, dataset):
        """ Add and return an existing dataset instance to the borehole.

            :param dataset: A precooked dataset with data
            :type dataset: subclassed from `pysiss.borehole.DataSet`
        """
        # Work out which attribute we should add the dataset to
        add_to_attr = self._type_to_attr[type(dataset)]

        # Add to the given attribute using the dataset name as a key
        getattr(self, add_to_attr)[dataset.name] = dataset
        return dataset

    def add_interval_dataset(self, name, from_depths, to_depths):
        """ Add and return a new IntervalDataSet

            :param name: The identifier for the new IntervalDataSet
            :type name: `string`
            :param from_depths: Interval start point down-hole depths in metres
                    from collar
            :type from_depths: iterable of numeric values
            :param to_depths: Interval end point down-hole depths in metres
                from collar
            :type to_depths: iterable of numeric values

            :returns: the new `pysiss.borehole.IntervalDataSet` instance.
        """
        return self.add_dataset(IntervalDataSet(name=name,
                                                from_depths=from_depths,
                                                to_depths=to_depths))

    def add_point_dataset(self, name, depths):
        """ Add and return a new PointDataSet.

            :param name: The identifier for the new PointDataSet
            :type name: `string`
            :param depths: Sample locations given as down-hole depths in metres
                    from collar
            :type depths: iterable of numeric values
            :returns: the new `pysiss.borehole.PointDataSet` instance.
        """
        return self.add_dataset(PointDataSet(name=name,
                                             depths=depths))

    def desurvey(self, depths, crs):
        """ Return the depths as three-dimensional points in the given
            coordinate reference system
        """
        raise NotImplementedError

    def add_merged_interval_dataset(self, name, source_name_a, source_name_b):
        """ Add a new merged interval dataset from the two sources
        """
        raise NotImplementedError

    def add_detail(self, name, values, property_type=None):
        """ Add a detail to this borehole object.

            :param name: An identifier for the detail
            :type name: string
            :param values: The data to add
            :type values: any Python object
            :param property_type: The property type of the detail, optional,
                   defaults to None
            :type property_type: pysiss.borehole.PropertyType
        """
        self.details.add_detail(name, values, property_type)


class Feature(id_object):

    """A point feature with properties but no spatial extent.

        Useful properties:
            depth - down-hole depth in metres
            properties - dict mapping property name to Property

        :param name: The identifier for the new PointDataSet
        :type name: `string`
        :param depth: Feature location given as down-hole depth in metres
                from collar
        :type depth: numeric value
    """

    def __init__(self, name, depth):
        super(Feature, self).__init__(name=name)
        self.name = name
        self.depth = depth
        self.properties = dict()

    def __repr__(self):
        """ String representation
        """
        info = 'Feature {0}: at {1} depth with {2} properties'
        return info.format(self.name, len(self.depth), len(self.properties))

    def add_property(self, property_type, values):
        """ Add a property to this feature.

            values - a single value or multiple values for a multivalued
                property
        """
        self.properties[property_type.name] = Property(property_type, values)

    def get_property_names(self):
        """ Return the names of the available properties for this feature
        """
        return self.properties.keys()


class CoordinateReferenceSystem(object):

    """System for describing a spatial location as a tuple of real numbers."""

    def __init__(self):
        raise NotImplementedError


class Survey(object):

    """ The spatial shape of the borehole path in three dimensions from the
        collar.

        Used to convert a sequence of down-hole depths into a sequence of
        three-dimensional points in some coordinate reference system.
    """

    def __init__(self):
        raise NotImplementedError


class OriginPosition(id_object):

    """Representation of borehole origin position in terms of latitude,
       longitude, and elevation.
    """

    def __init__(self, latitude, longitude, elevation, property_type=None):
        super(OriginPosition, self).__init__(name=str((latitude,
                                                       longitude,
                                                       elevation,
                                                       property_type)))
        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation
        self.property_type = property_type

    def __repr__(self):
        """ String representation
        """
        info = "latitude {0}, longitude {1}, elevation {2}"
        info = info.format(self.latitude, self.longitude, self.elevation)
        if self.property_type is not None:
            info = (info + ", {0}").format(self.property_type)
        return info


class BoreholeDetails(Details):

    """ Class to store details about drilling a Borehole
    """

    detail_type = detail_type('BoreholeDetail', 'name values property_type')



""" file:   unmarshal.py (pysiss.vocabulary)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Monday 25 August, 2014

    description: Wrapper functionality for unmarshalling XML elements
"""

from .namespaces import shorten_namespace, expand_namespace
from .gml import unmarshallers as gml
from .gsml import unmarshallers as gsml
from .erml import unmarshallers as erml

from lxml import etree

UNMARSHALLERS = {}
UNMARSHALLERS.update(gml.UNMARSHALLERS)
UNMARSHALLERS.update(gsml.UNMARSHALLERS)
UNMARSHALLERS.update(erml.UNMARSHALLERS)


def unmarshal(elem):
    """ Unmarshal an lxml.etree.Element element

        If there is no unmarshalling function available, this just returns the
        lxml.etree element.
    """
    tag = shorten_namespace(elem.tag)
    unmarshal = UNMARSHALLERS.get(tag)
    if unmarshal:
        return unmarshal(elem)
    else:
        return None


def unmarshal_all(filename, tag='gsml:MappedFeature'):
    """ Unmarshall all instances of a tag from an xml file
        and return them as a list of objects
    """
    tag = expand_namespace(tag)
    results = []
    with open(filename, 'rb') as fhandle:
        try:
            context = iter(etree.iterparse(fhandle, events=('end',), tag=tag))
            for event, elem in context:
                results.append(unmarshal(elem))
        except etree.XMLSyntaxError:
            pass
    return results



""" file:   borehole_generator.py (pysiss.borehole)
    author: David Benn
            CSIRO IM&T Scientific Computing Data Processing Services
    date:   18 February 2014

    description: Borehole object creation from SISS GeoSciML metadata.

Notes: 
o If the expectation is that only one borehole element should be 
  found, should we raise an exception if there is more than one?
  Or, should we check that the name is the same as the borehole's 
  identifier? Since the caller can currently pass whatever name 
  he/she desires, that test may fail.
"""

import re
import xml.etree.ElementTree
from datetime import datetime
from pint import UnitRegistry

from ..properties import PropertyType
from ..borehole import Borehole, OriginPosition

# General namespace URIs for GeoSciML
NS = {'gsml': 'urn:cgi:xmlns:CGI:GeoSciML:2.0',
      'gsmlbh': 'http://xmlns.geosciml.org/Borehole/3.0',
      'xlink': 'http://www.w3.org/1999/xlink'}

# GeoSciML version dependent gml namespace URIs
GML_NS = {'gsml': 'http://www.opengis.net/gml',
          'gsmlbh': 'http://www.opengis.net/gml/3.2'}

# GeoSciML version dependent shape namespace URIs
SHAPE_NS = {'gsml': 'http://www.opengis.net/sampling/1.0',
            'gsmlbh': 'http://www.opengis.net/samplingSpatial/2.0'}
              
class SISSBoreholeGenerator:

    """ Spatial Information Services Stack borehole generator class.
        Creates borehole objects from GeoSciML input.

        The following GeoSciML namespace variations are handled:

            xmlns:gsml => urn:cgi:xmlns:CGI:GeoSciML:2.0
            xmlns:gsmlbh => http://xmlns.geosciml.org/Borehole/3.0
            
        Borehole details that are in common across GeoSciML 2.0 and 3.0
        are extracted.
    """

    def __init__(self):
        """ Construct a SISS borehole generator instance.
        """
        self.unit_reg = UnitRegistry() 

        self.ns_key = None

        self.geosciml_handlers = {}
        self.geosciml_handlers['gsml'] = self._add_gsml_borehole_details
        self.geosciml_handlers['gsmlbh'] = self._add_gsmlbh_borehole_details

        self.whitespace_pattern = re.compile(r'\s+')
                
        self.borehole = None
    
    def geosciml_to_borehole(self, name, geo_source):
        """ Given a GeoSciML scanned borehole URL, return a Borehole object
            initialised with origin position and borehole details. In the case
            where there is more than one Borehole element, the first will be
            returned.

            :param name: The name to assign to the Borehole object
            :type name: string
            :param geo_source: A file-like object opened from a GeoSciML
                scanned borehole URL
            :type geo_source: file-like object
            :returns: a Borehole object initialised with origin position and
                borehole details
        """
        if geo_source is not None:
            geo_tree = xml.etree.ElementTree.parse(geo_source)
            borehole_elts = self._get_borehole_elts(geo_tree)

            if len(borehole_elts) != 0:
                self.borehole = Borehole(name=name,
                            origin_position=self._location(borehole_elts[0]))
    
                self._add_borehole_details(borehole_elts[0])

        return self.borehole

    def _get_borehole_elts(self, geo_tree):
        """ Return a list of GeoSciML Borehole elements taking into account
            tag namespace variations.

            :param geo_tree: a GeoSciML element tree
            :type geo_tree: xml.etree.ElementTree
            :returns: a list of Borehole elements
        """
        boreholes = []
        for ns_prefix in ['gsml', 'gsmlbh']:
            boreholes = geo_tree.findall('.//{' + NS[ns_prefix] + '}Borehole')
            if len(boreholes) != 0:
                self.ns_key = ns_prefix
                break

        return boreholes

    def _location(self, borehole_elt):
        """Find the GeoSciML 2.0 or 3.0 borehole position (lat/lon) and 
           elevation and return an OriginPosition instance.
           
        :param borehole_elt: A GeoSciML 2.0 or 3.0 Borehole element
        :type borehole_elt: Element
        :returns: an OriginPosition instance (or None, if not found)
        """
        origin_position = None
        
        latlon_xpath = './/{{{0}}}location/{{{1}}}Point/{{{2}}}pos'
        latlon = _element_text(borehole_elt,
                               latlon_xpath.format(NS[self.ns_key],
                                                   GML_NS[self.ns_key],
                                                   GML_NS[self.ns_key]))
        if latlon is not None:
            (lat, lon) = latlon.split(' ')
         
            elevation_xpath = './/{{{0}}}elevation[@uomLabels]'
            elevation_elt = \
                borehole_elt.find(elevation_xpath.format(NS[self.ns_key]))
            
            if elevation_elt is not None:
                elevation_units = \
                    self.unit_reg[elevation_elt.attrib['uomLabels']]
            else:
                elevation_units = None
                
            if self.ns_key == 'gsml':
                property_type = \
                    self._gsml_location_property(borehole_elt,
                                                 elevation_units)
            else:
                property_type = self._gsmlbh_location_property(borehole_elt)
                
            origin_position = \
                OriginPosition(latitude=float(lat) * self.unit_reg.degree,
                    longitude=float(lon) * self.unit_reg.degree,
                    elevation=float(elevation_elt.text) * elevation_units,
                    property_type=property_type)

        return origin_position

    def _gsml_location_property(self, borehole_elt, units):
        """Return a GeoSciML 2.0 location (elevation) property object.
        
        :param borehole_elt: A GeoSciML 2.0 Borehole element
        :type borehole_elt: Element
        :param units: elevation units
        :type units: A Pint elevation unit (e.g. meters)
        :returns: a location (elevation) property (or None, if not found)
        """
        property_type = None
        
        elevation_desc_xpath = './/{{{0}}}elevation[@axisLabels]'
        elevation_elt = \
            borehole_elt.find(elevation_desc_xpath.format(NS[self.ns_key]))
        if elevation_elt is not None:
            elevation_axis_desc = \
                'elevation: {0}'.format(elevation_elt.attrib['axisLabels'])
            property_type = PropertyType(name='origin position elevation',
                                         long_name='origin position elevation',
                                         description=elevation_axis_desc,
                                         units=units)
            
        return property_type
    
    def _gsmlbh_location_property(self, borehole_elt):
        """Return a GeoSciML 3.0 location (description) property object.
        
        :param borehole_elt: A GeoSciML 2.0 Borehole element
        :type borehole_elt: Element
        :returns: a location (description) property (or None, if not found)
        """
        description_xpath = \
            './/{{{0}}}location/{{{1}}}Point/{{{2}}}description'
            
        description_text = _element_text(borehole_elt,
                         description_xpath.format(NS[self.ns_key],
                                                  GML_NS[self.ns_key],
                                                  GML_NS[self.ns_key]))
        
        description_text = 'description: {0}'.format(description_text)
        
        return PropertyType(name='origin position',
                            long_name='origin position',
                            description=description_text)
    
    def _add_borehole_details(self, borehole_elt):
        """ Add borehole details.

            This top-level method calls more specific methods to add
            borehole details.

            :param borehole_elt: A GeoSciML Borehole element
            :type borehole_elt: Element
        """
        details_elt = \
        borehole_elt.find('.//{{{0}}}BoreholeDetails'.format(NS[self.ns_key]))
        if details_elt is not None:
            return self.geosciml_handlers[self.ns_key](borehole_elt,
                                                       details_elt)

    def _add_gsml_borehole_details(self, borehole_elt, details_elt):
        """Add borehole details from a GeoSciML 2.0 Borehole or
            BoreholeDetails element.
        
        :param borehole_elt: A GeoSciML 2.0 Borehole element
        :type borehole_elt: Element
        :param details_elt: A GeoSciML 2.0 BoreholeDetails element
        :type details_elt: Element
        """
        # Driller
        self._add_driller(details_elt)

        # Drilling method
        drilling_method = \
            _element_text(details_elt,
                          './/{{{0}}}drillingMethod'.format(NS[self.ns_key]))
        self.borehole.add_detail('drilling method', drilling_method)
        
        # Date of drilling
        date_of_drilling = \
            _element_text(details_elt,
                          './/{{{0}}}dateOfDrilling'.format(NS[self.ns_key]))
        year, month, day = date_of_drilling.split('-')
        date = datetime(year=int(year), month=int(month), day=int(day))
        self.borehole.add_detail('date of drilling', date)

        # Borehole start point
        start_point = \
            _element_text(details_elt,
                          './/{{{0}}}startPoint'.format(NS[self.ns_key]))
        self.borehole.add_detail('start point', start_point)
        
        # Borehole inclination type
        inclination_type = \
            _element_text(details_elt,
                          './/{{{0}}}inclinationType'.format(NS[self.ns_key]))
        self.borehole.add_detail('inclination type', inclination_type)
        
        # Borehole shape
        # Note: This is a child of the Borehole element rather than
        #       BoreholeDetails. 
        shape_xpath = './/{{{0}}}shape/{{{1}}}LineString/{{{2}}}posList'
        shape = \
            _element_text(borehole_elt,
                          shape_xpath.format(SHAPE_NS[self.ns_key],
                                             GML_NS[self.ns_key],
                                             GML_NS[self.ns_key]))
        shape_list = [float(x) for x in self.whitespace_pattern.split(shape)]
        self.borehole.add_detail('shape', shape_list)
        
        # Borehole cored interval
        cored_interval_xpath = \
            './/{{{0}}}coredInterval/{{{1}}}Envelope[@uomLabels]'
        cored_interval_elt = \
            details_elt.find(cored_interval_xpath.format(NS[self.ns_key],
                                                         GML_NS[self.ns_key]))
        cored_interval_units = \
            self.unit_reg[cored_interval_elt.attrib['uomLabels']]
            
        cored_interval_lower_corner = \
            _element_text(cored_interval_elt,
                          './/{{{0}}}lowerCorner'.format(GML_NS[self.ns_key]))
        cored_interval_upper_corner = \
            _element_text(cored_interval_elt,
                          './/{{{0}}}upperCorner'.format(GML_NS[self.ns_key]))
            
        lower_corner = float(cored_interval_lower_corner) * cored_interval_units
        upper_corner = float(cored_interval_upper_corner) * cored_interval_units
        envelope_dict = {'lower corner': lower_corner,
                         'upper corner': upper_corner}
        
        # Question: How useful is the property here in fact if we have units
        #           for each value?
        self.borehole.add_detail('cored interval', envelope_dict,
                            PropertyType(name='envelope',
                                         long_name='cored interval envelope',
                                         description='cored interval envelope '
                                                     'lower and upper corner',
                                         units=cored_interval_units))
                       
    def _add_gsmlbh_borehole_details(self, borehole_elt, details_elt):
        """Add borehole details from a GeoSciML 3.0 Borehole or
           BoreholeDetails element.
        
        :param borehole_elt: A GeoSciML 3.0 Borehole element
        :type borehole_elt: Element
        :param details_elt: A GeoSciML 3.0 BoreholeDetails element
        :type details_elt: Element
        """

        # Driller
        self._add_driller(details_elt)
        
        # Drilling method
        # Note:  This is a child of the Borehole element rather than
        #        BoreholeDetails. 
        drilling_method_xpath = ('.//{{{0}}}downholeDrillingDetails'
                                 '/{{{1}}}DrillingDetails'
                                 '/{{{2}}}drillingMethod'). \
                                 format(NS[self.ns_key],
                                        NS[self.ns_key],
                                        NS[self.ns_key])
                                                        
        drilling_method_attrib_xpath = '{{{0}}}title'.format(NS['xlink'])
        drilling_method = _element_attrib(borehole_elt,
                                          drilling_method_xpath,
                                          drilling_method_attrib_xpath)
        self.borehole.add_detail('drilling method', drilling_method)

        # Date of drilling
        # Note: Both start and end time are available; currently extracting
        #       only start time.
        date_of_drilling_xpath = ('.//{{{0}}}dateOfDrilling/{{{1}}}TimePeriod'
                            '/{{{2}}}begin/{{{3}}}TimeInstant'
                            '/{{{4}}}timePosition').format(NS[self.ns_key],
                                                           GML_NS[self.ns_key],
                                                           GML_NS[self.ns_key],
                                                           GML_NS[self.ns_key],
                                                           GML_NS[self.ns_key])
        date_of_drilling = _element_text(details_elt,
                                         date_of_drilling_xpath)
        year, month, day = date_of_drilling.split('-')
        date = datetime(year=int(year), month=int(month), day=int(day))
        self.borehole.add_detail('date of drilling', date)

        # Borehole start point
        start_point_xpath = './/{{{0}}}startPoint'.format(NS[self.ns_key])
        start_point_attrib_xpath = '{{{0}}}title'.format(NS['xlink'])
        start_point = _element_attrib(details_elt, start_point_xpath,
                                      start_point_attrib_xpath)
        self.borehole.add_detail('start point', start_point)

        # Borehole inclination type
        inclination_type_xpath = \
            './/{{{0}}}inclinationType'.format(NS[self.ns_key])
        inclination_type_attrib_xpath = '{{{0}}}title'.format(NS['xlink'])
        inclination_type = _element_attrib(details_elt,
                                            inclination_type_xpath,
                                            inclination_type_attrib_xpath)
        self.borehole.add_detail('inclination type', inclination_type)
        
        # Borehole shape
        # Notes: 
        # o This is a child of the Borehole element rather than BoreholeDetails.
        # o Currently chooses the first one (if more than one exists).
        shape_xpath = './/{{{0}}}shape/{{{1}}}CompositeCurve' \
                      '/{{{2}}}curveMember/{{{3}}}LineString/{{{4}}}posList'
        shape = \
            _element_text(borehole_elt,
                          shape_xpath.format(SHAPE_NS[self.ns_key],
                                             GML_NS[self.ns_key],
                                             GML_NS[self.ns_key],
                                             GML_NS[self.ns_key],
                                             GML_NS[self.ns_key]))
        shape_list = [float(x) for x in self.whitespace_pattern.split(shape)]
        self.borehole.add_detail('shape', shape_list)
        
        # Borehole cored interval
        # Note: No units; haven't used a PropertyType here.
        cored_interval_xpath = ('.//{{{0}}}downholeDrillingDetails'
                                 '/{{{1}}}DrillingDetails'
                                 '/{{{2}}}interval/{{{3}}}LineString'
                                 '/{{{4}}}posList'). \
                                 format(NS[self.ns_key],
                                        NS[self.ns_key],
                                        NS[self.ns_key],
                                        GML_NS[self.ns_key],
                                        GML_NS[self.ns_key])
        
        cored_interval = _element_text(borehole_elt,
                                       cored_interval_xpath)
        cored_interval_list = self.whitespace_pattern.split(cored_interval)
        lower_corner = float(cored_interval_list[0])
        upper_corner = float(cored_interval_list[1])
        envelope_dict = {'lower corner': lower_corner,
                         'upper corner': upper_corner}
        self.borehole.add_detail('cored interval', envelope_dict)
    
    def _add_driller(self, details_elt):
        """Add borehole driller detail from a GeoSciML 3.0
           BoreholeDetails element.
        
        :param details_elt: A GeoSciML 3.0 BoreholeDetails element
        :type details_elt: Element
        """
        driller_xpath = './/{{{0}}}driller'.format(NS[self.ns_key])
        driller_attrib_xpath = '{{{0}}}title'.format(NS['xlink'])
        driller = _element_attrib(details_elt, driller_xpath,
                                  driller_attrib_xpath)
        self.borehole.add_detail('driller', driller)

def _element_text(element, xpath_str):
    """Look for and return the detail corresponding to the text 
       of the child element found by the specified XPath search
       on the supplied element.
       
    :param element: An XML element
    :type element: Element
    :param xpath_str: An XPath string.
    :type xpath_str: string
    :returns: child text
    """
    result = element.find(xpath_str)
   
    if result is not None:
        text = result.text
    else:
        text = None
    
    return text

def _element_attrib(element, elt_xpath_str, attr_xpath_str):
    """Look for and return the detail corresponding to the attribute 
       text of the child element found by the specified XPath search
       on the supplied element.
       
    :param element: An XML element
    :type element: Element
    :param elt_xpath_str: An XPath string to find an element.
    :type elt_xpath_str: string
    :param attr_xpath_str: An XPath string to find an attribute.
    :type attr_xpath_str: string
    :returns: attribute text
    """
    result = element.find(elt_xpath_str)
   
    if result is not None:
        text = result.attrib[attr_xpath_str]
    else:
        text = None
    
    return text



""" file: projection.py (pysiss.utilities)

    description: Projection utilities
"""

# from osgeo import ogr
# from shapely.wkb import loads


# def project(geom, to_epsg=900913, from_epsg=4326):
#     """utility function to do quick projection with ogr,
#     to and from shapely objects

#         >>> from shapely.geometry import LineString
#         >>> l = LineString([[-121, 43], [-122, 42]])
#         >>> lp = project(l, from_epsg=4326, to_epsg=26910)
#         >>> lp.wkt
#         'LINESTRING (663019.0700828594854102 4762755.6415722491219640,
#             582818.0692490270594135 4650259.8474613213911653)'

#         Stolen from http://hackmap.blogspot.com.au/2008/03/ogr-python-projection.html
#     """

#     to_srs = ogr.osr.SpatialReference()
#     to_srs.ImportFromEPSG(to_epsg)

#     from_srs = ogr.osr.SpatialReference()
#     from_srs.ImportFromEPSG(from_epsg)

#     ogr_geom = ogr.CreateGeometryFromWkb(geom.wkb)
#     ogr_geom.AssignSpatialReference(from_srs)

#     ogr_geom.TransformTo(to_srs)
#     return loads(ogr_geom.ExportToWkb())



""" file: detrend.py (pysiss.borehole.analysis)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Utilities to low-pass filter regularly spaced data in a
    PointDataSet instance (i.e. detrending).
"""

import numpy
import scipy.optimize


def demean(data, axis=0):
    """ Return data minus its mean along the specified axis.

        :param data: Array of data
        :type data: `numpy.ndarray`
        :param axis: Axis along which to take means
        :type axis: int
        :returns: Demeaned data
    """
    data = numpy.asarray(data)
    if axis == 0 or axis is None or data.ndim <= 1:
        return data - data.mean(axis)
    else:
        index = [slice(None)] * data.ndim
        index[axis] = numpy.newaxis
        return data - data.mean(axis)[index]


# Detrend helper functions
def _detrend_mean(data):
    """ Detrend a signal by subtracting the mean.

        This function is internal to cwavelets.utilities and should not be
        called outside of it. See cwavelets.utilities.detrend instead.

        Data is modified in place.

        :param data: the one-dimensional input data
        :type data: `numpy.array`
        :returns: None (data modified in place)
    """
    data -= data.mean()


def _detrend_function(data, func, param_guess):
    """ Detrend a signal using the given function.

        Calculates the best fit trend function using `numpy.linalg.leastsq`
        and subtracts it from the data. Data is modified in place.

        This function is internal to cwavelets.utilities and should not be
        called outside of it. See cwavelets.utilities.detrend instead.

        :param data: the one-dimensional input data
        :type data: `numpy.array`
        :param func: specify a model function to use in the
            detrending.
        :param param_guess: An initial guess at the parameters in the model.
            Usually an array of ones, or an array with the highest polynomial
            factor set to one and the rest to zero works well. A `ValueError`
            will be raised if `func` is specified but not `param_guess`.
        :type param_guess: `numpy.ndarray`
        :returns: None (data modified in place)
    """
    # Rescale the data to lie between 0 and 1
    drange, dmean = numpy.max(data) - numpy.min(data), data.mean()
    data_scaled = (data - dmean) / drange

    # We define the dataset to just be the unit interval which is fine for most
    # detrending things, when the actual parameter values don't matter
    dataset = numpy.linspace(0, 1, len(data))
    residuals = lambda p, y, x: y - func(p, x)

    # Perform least-squares fit
    param_guess = numpy.asarray(param_guess)
    param_lsq, flag = scipy.optimize.leastsq(
        residuals, param_guess, args=(data_scaled, dataset))
    if flag not in [1, 2, 3, 4]:
        raise ValueError("Least-squares routine failed to converge")
    data -= drange * func(param_lsq, dataset) + dmean
    return None


BUILTIN_TRENDS = {
    'none': lambda data: None,  # Not sure why you'd use this
    'mean': _detrend_mean,
    'linear': lambda data: _detrend_function(
        data,
        func=lambda b, x: b[1] * x + b[0],
        param_guess=[1, 0]),
    'quadratic': lambda data: _detrend_function(
        data,
        func=lambda b, x: b[2] * x ** 2 + b[1] * x + b[0],
        param_guess=[1, 0, 0]),
    'cubic': lambda data: _detrend_function(
        data,
        func=lambda b, x: b[3] * x ** 3 + b[2] * x ** 2 + b[1] * x + b[0],
        param_guess=[1, 0, 0, 0])
}


def detrend(data, trend=None, func=None, param_guess=None):
    r""" Detrend a data array in-place using the given method

        The behavior of the function depends on the trend supplied:
            -   If `trend` is `none`, no detrending is done
            -   if `trend` is `mean`, the data mean is subtracted from each
                data point.
            -   If `trend` is one of `linear`, `quadratic` or `cubic`, the
                data is detrended using best least-squares fit to the relevant
                model.

        Alternatively you can supply your own function for detrending using
        the func argument. This model function should take a one-dimensional
        numpy array, plus a vector of variables which can be fit to the data
        using `scipy.optimize.leastsq`. Note that specifying both `trend` and
        `func` when calling this function will raise a ValueError.

        If no extra parameters are specified, this function performs linear
        detrending.

        Internally, the input data are rescaled to the unit inteval to improve
        the detrending. This might cause a problem for cases where the range
        of the data is very small relative to machine precision.

        :param data: the one-dimensional input data
        :type data: `numpy.array`
        :param trend: Optional, use a builtin trend model. One of 'none',
            'mean', 'linear', 'quadratic' or 'cubic'. Defaults to 'linear' if
            `func` is not specified.
        :type trend: str
        :param func: Optional, specify a model function to use in the
            detrending. The function should be of the form :math:`f(p, x)`
            where :math:`x` is some dataset parameter, and :math:`p` is a
            paramter vector for the model. See `scipy.optimize.leastsq` for
            more details on the form of this function.
        :param param_guess: An initial guess at the parameters in the model.
            Usually an array of ones, or an array with the highest polynomial
            factor set to one and the rest to zero works well. A `ValueError`
            will be raised if `func` is specified but not `param_guess`.
        :type param_guess: `numpy.ndarray`
        :returns: None (data modified in place)
    """
    # Deal with default arguments
    if trend is not None and func is not None:
        raise ValueError("You should only specify one of the trend or func "
                         "keyword arguments to detrend.")
    elif func is not None:
        if param_guess is None:
            raise ValueError("You need to specify starting parameters for "
                             "your trend model.")
        return _detrend_function(data, func, param_guess)

    # If we're here, we are going to use a builtin funciton.
    default = 'linear'
    if (trend is not None) and (trend not in BUILTIN_TRENDS.keys()):
        raise ValueError("trend should be one of {0}"
                         .format(BUILTIN_TRENDS.keys()))
    return BUILTIN_TRENDS[trend or default](data)



#!/usr/bin/env python
""" file:   __init__.py (pysiss.borehole.analysis)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    email:  jesse.robertson@csiro.au
    date:   Wednesday May 1, 2013

    description: Initialisation of the pysiss.borehole.modifiers module.
"""

from .regularizer import ReSampler, unique
from .detrend import detrend, demean


""" file: regularizer.py (pysiss.borehole.analysis)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Utilities to generate regularly spaced data from irregular
        grids.
"""

import numpy
import scipy.interpolate

# Reinterpolate functions
def unique(array, return_index=True, sort_method='heapsort', eqtest=None):
    """ Find the unique elements of an array.

        Returns the sorted unique elements of an array. There are two optional
        outputs in addition to the unique elements: the indices of the input
        array that give the unique values, and the indices of the unique array
        that reconstruct the input array.

        :param array: Input array. This will be flattened if it is not already 1-D.
        :type array: array_like
        :param return_index: If True, also return the indices of `array` that result in the unique array. Optional, defaults to True.
        :type return_index: bool
        :param sort_method: The method used to sort the array. Defaults to 'heapsort', which has the best worst-case running time for pre-sorted arrays.
        :type sort_method: one of 'quicksort', 'mergesort' or 'heapsort'
        :param eqtest: Equality test function used to determine whether two elements are equal. The test function should take two numpy arrays and return a numpy bool array with the result of an elementwise test. Defaults to `lambda a, b: a ==b`.
        :type eqtest: numpy.ufunc
        :returns: The sorted unique values, and optionally the indices of the first occurrences of the unique values in the (flattened) original array (if `return_index` is True).
    """
    # Sort out array input
    try:
        array = array.flatten()
    except AttributeError:
        array = numpy.asanyarray(array).flatten()
    if array.size == 0:
        return array, numpy.empty(0, numpy.bool)

    # Set default equality test
    if eqtest is None:
        eqtest = lambda a, b: a == b

    # Do sort
    perm = array.argsort(kind=sort_method)
    aux = array[perm]

    # Check for inequality (i.e. we want to mask values with False in the mask
    # array for which eqtest is true)
    neqflag = numpy.concatenate(([True],
        numpy.logical_not(eqtest(aux[1:], aux[:-1]))))
    if return_index:
        return aux[neqflag], perm[neqflag]
    else:
        return aux[neqflag]


class ReSampler(scipy.interpolate.InterpolatedUnivariateSpline):
    """ Resamples a dataset over a given dataset onto a regularly
        gridded dataset.

        On initialisation, a spline fit to the signal is generated. Subsequent
        calls to the ReSampler instance will use this spline to generate the
        resampled arrays.

        This is really just a wrapper around `scipy.interpolate.
        InterpolatedUnivariateSpline` which adds some cruft to generate a
        uniform grid. If you want a non-uniform grid then you might like to
        take a look at that function.

        :param dataset: the locations of each signal value in the dataset.
            Optional, if `spline` is specified then these signal will be
            ignored.
        :type dataset: `numpy.ndarray`
        :param signal: the signal evaluated at a given set of points. If
            `spline` is specified then these signal will be ignored.
        :type signal: `numpy.ndarray`
        :param order: the interpolation order for the spline. Optional,
            defaults to 3 (giving cubic interpolation)
        :type order: `int`
    """

    def __init__(self, dataset, signal, order=3):
        """ Initialise the ReSampler instance
        """
        # Check inputs
        if dataset is None or signal is None:
            raise ValueError("You must specify both signal and locations to"
                " resample signal.")
        elif len(dataset) != len(signal):
            raise ValueError("DataSet and value arrays are different lengths.")
        self.dataset, self.signal = dataset, signal
        self.order = order

        # Data points must be increasing, perform a sort if not
        sorted_signal, sorted_index = \
            unique(self.dataset, sort_method='heapsort',
                eqtest=lambda a, b: (a - b) ** 2 / numpy.abs(a) <= 1e-12)
        sorted_dataset = self.dataset[sorted_index]
        sorted_signal = self.signal[sorted_index]

        # Get dataset bounds
        self.dataset_bounds = (sorted_dataset[0], sorted_dataset[-1])

        # Initialise underlying Spline instance
        super(ReSampler, self).__init__(
            x=sorted_dataset, y=sorted_signal, w=None,
            bbox=self.dataset_bounds, k=self.order)

    def resample(self, nsamples, dataset_bounds=None, derivative=0):
        """ Generate a resampled dataset.

            The locations of the resampled points are uniformly distributed over the range specified in `dataset_bounds`, which defaults to the range of dataset values supplied.

            :param nsamples: number of resampled points
            :type nsamples: int
            :param dataset_bounds: the range to resample over, given as a pair `(dmin, dmax)`. Optional - if the dataset_values array not is specified, this defaults to the max and min of the input data.
            :type dataset_bounds: 2-tuple of int
            :param derivative: The order of derivative to return. `derivative=0` returns the value at x, `derivative=1` the first derivative and so on...
            :type derivative: int
            :returns: Two `numpy.ndarray` of length nsamples containing the resampled dataset and the dataset grid
        """
        if dataset_bounds is None:
            dataset_bounds = self.dataset_bounds
        resampled_dataset = numpy.linspace(dataset_bounds[0],
            dataset_bounds[1], nsamples)
        resampled_data = super(ReSampler, self).__call__(
            x=resampled_dataset, nu=derivative)
        return resampled_dataset, resampled_data



""" file:   vector.py (pysiss.coverage)
    author: Jess Robertson
            CSIRO Minerals Resources Flagship
    date:   Monday 25 August, 2014

    desription: Implementation of classes for vector coverage data
"""

from ..utilities import id_object
from ..metadata import MetadataRegistry


class MappedFeature(id_object):

    """ Class containing vector GIS data.

        Corresponds roughly to gsml:MappedFeatures
    """

    md_registry = MetadataRegistry()

    def __init__(self, shape, projection, specification, ident=None, **kwargs):
        super(MappedFeature, self).__init__(name='mapped_feature')
        self.ident = ident or self.uuid

        # Store some info on the shape
        self.shape = shape
        self.projection = projection
        self.centroid = self.shape.representative_point()

        # Store other metadata
        for attrib, value in kwargs.items():
            setattr(self, attrib, value)
        self.specification = specification
        self.type = self.md_registry[self.specification].type

    def __repr__(self):
        """ String representation
        """
        info = 'MappedFeature {0} somewhere near {1} contains '
        info_str = info.format(self.ident, self.centroid)
        return info_str

    def reproject(self, new_projection):
        """ Reproject the shape to a new projection.

            :param new_projection: The identifier for the new projection
            :type new_projection: int
        """
        raise NotImplementedError
        # self.shape = type(self.shape)(project(self.shape.positions))

    @property
    def metadata(self):
        """ Return the metadata associated with the MappedFeature
        """
        return self.md_registry[self.specification]



""" file: __init__.py (pysiss.borehole.properties)
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date: Sunday November 10, 2013

    description: Imports for pysiss.borehole.properties
"""

from .property import Property
from .property_type import PropertyType


""" file:   test_nvcl.py
    author: Jess Robertson
            CSIRO Earth Science and Resource Engineering
    date:   Tuesday 13 May, 2014

    description: Tests for NVCL importer
"""

import unittest
import pysiss.webservices.nvcl as nvcl


class TestNVCLEndpointRegistry(unittest.TestCase):

    def setUp(self):
        self.registry = nvcl.NVCLEndpointRegistry()

    def test_noclobber_ident(self):
        """ Check that existing idents are not clobbered
        """
        urls = ('a/url', 'b/url', 'c/url')
        self.assertRaises(KeyError, self.registry.register, 'CSIRO', *urls)

    def test_update(self):
        """ Check that endpoint data is updated
        """
        # Add a new key and then update it
        keys = ('wfsurl', 'dataurl', 'downloadurl')
        urls = ('a/url', 'b/url', 'c/url')
        self.registry.register('myendpoint', *urls)
        for url, key in zip(urls, keys):
            self.assertEquals(url, self.registry['myendpoint'][key])

        # Update the keys
        updated_urls = ('d/url', 'e/url', 'f/url')
        self.registry.register('myendpoint', *updated_urls,
                               update=True)
        for url, key in zip(updated_urls, keys):
            self.assertEquals(url, self.registry['myendpoint'][key])

        # Remove bogus data
        del self.registry['myendpoint']

    def test_singleton(self):
        """ Check that only one instance of the registry
            is instantiated
        """
        self.assertEquals(nvcl.NVCLEndpointRegistry(),
                          self.registry)

    def tearDown(self):
        # Remove test changes to registry
        try:
            del self.registry['myendpoint']
        except KeyError:
            # It's already deleted, or never existed anyway
            pass


class TestNVCLImporter(unittest.TestCase):

    """ Test NVCLImporter class
    """

    def setUp(self):
        # Get a pointer to the NVCL endpoint registry
        self.registry = nvcl.NVCLEndpointRegistry()

        # Make importers for each default endpoint
        self.importers = {}
        for ept in self.registry.keys():
            self.importers[ept] = nvcl.NVCLImporter(ept)

    def test_usage_1(self):
        """ Test some sample usage using the GSWA endpoint
        """
        bh_info = self.importers['GSWA'].get_borehole_idents_and_urls()

        for ident, url in dict(bh_info).items():
            self.assertTrue(ident != '')
            self.assertTrue(url != '')

    def test_usage_2(self):
        """ Test some sample usage using the GSWA endpoint
        """
        bh_idents = self.importers['GSWA'].get_borehole_idents()

        # Get the datasets assocated with the borehole ident
        datasets = {}
        for ident in bh_idents:
            datasets[ident] = \
                self.importers['GSWA'].get_dataset_idents(ident)

    def test_usage_3(self):
        """ Test some sample usage using the GSWA endpoint
        """
        self.importers['GSWA'].get_borehole('PDP2C', get_analytes=True)



#!/usr/bin/env python
""" file: setup.py (pysiss)
    author: Jess Robertson, CSIRO Earth Science and Resource Engineering
    date: Wednesday 1 May, 2013

    description: Setuptools installer script for pysiss.
"""

from setuptools import setup, find_packages
import os

# Get requirements from requirements.txt file
with open('requirements.txt') as fhandle:
    REQUIREMENTS = map(lambda l: l.strip('\n'), fhandle.readlines())


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

## PACKAGE INFORMATION
setup(
    # Metadata
    name='pysiss',
    version='0.0.3',
    description='A pythonic interface to Spatial Information Services Stack '
                '(SISS) services',
    long_description=read('README.rst'),
    author='Jess Robertson',
    author_email='jesse.robertson@csiro.au',
    url='http://github.com/pysiss/pysiss',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: XML'
    ],

    # Dependencies
    install_requires=[
        'matplotlib>=1.0',
        'numpy>=1.6',
        'scipy>=0.9',
        'OWSLib>=0.8',
        'lxml',
        'simplejson>=3.0',
        'pandas>=0.10',
        'shapely',
        'requests',
        'pint'
    ],

    # Contents
    packages=find_packages(exclude=['test*']),
    package_data={
        'pysiss.vocabulary.resources': ['*']
    },
    test_suite='tests'
)



""" file:   interval_dataset.py (pysiss.borehole.datasets)
    author: Jess Robertson & Ben Caradoc-Davies
            CSIRO Earth Science and Resource Engineering
    date:   Sunday November 10, 2013

    description: DataSet for interval data (data defined over an interval in
        the borehole).

    An IntervalDataSet is is a sequence of borehole segments each having a
    single value for each property; this value is taken to be the same across
    the entire length of the interval. IntervalSampling can be merged to form a
    new IntervalDataSet that has the intervals whose boundaries are the union
    of the boundaries of the source IntervalSampling. An IntervalDataSet can be
    interpolated onto a PointDataSet.

    Intervals must be in depth order and not overlap, but there might
    be gaps between intervals.
"""

from .dataset import DataSet
from .point_dataset import PointDataSet

import numpy
import pandas


class IntervalDataSet(DataSet):

    """ IntervalDataSet contains data which is defined over some depth
        interval.

        An IntervalDataSet is is a sequence of borehole segments each having a
        single value for each property; this value is taken to be the same
        across the entire length of the interval. IntervalSampling can be
        merged to form a new IntervalDataSet that has the intervals whose
        boundaries are the union of the boundaries of the source
        IntervalSampling. An IntervalDataSet can be interpolated onto a
        PointDataSet.

        Intervals must be in depth order and not overlap, but there might
        be gaps between intervals.

        :param name: identifier for the dataSet
        :type name: string
        :param from_depths: interval start point down-hole depths in metres
            from collar
        :type from_depths: iterable
        :param to_depths: interval end point down-hole depths in metres from
            collar
        :type to_depths: iterable
        :param details: The metadata associated with the dataset. Optional,
            defaults to None.
        :type details: pysiss.borehole.dataset.DatasetDetails
    """

    def __init__(self, name, from_depths, to_depths, details=None):
        super(IntervalDataSet, self).__init__(
            name, len(from_depths), details=details)
        from_depths = numpy.asarray(from_depths)
        to_depths = numpy.asarray(to_depths)
        assert len(from_depths) == len(to_depths), \
            "from_ and to_depths must be same length"
        assert all(numpy.gradient(from_depths) > 0), \
            "from_depths must be monotonically increasing"
        assert all(numpy.gradient(to_depths) > 0), \
            "to_depths must be monotonically increasing"
        assert all(to_depths - from_depths > 0), \
            "intervals must have positive length"
        assert all(to_depths[:-1] <= from_depths[1:]), \
            "intervals must not overlap"
        self.from_depths = from_depths
        self.to_depths = to_depths

    def __repr__(self):
        info = 'IntervalDataSet {0}: with {1} depth intervals and {2} '\
               'properties'
        return info.format(self.name, len(self.from_depths),
                           len(self.properties))

    def get_interval(self, from_depth, to_depth, dataset_name=None):
        """ Return the data between the given depths as as new IntervalDataSet

            Only intervals completely contained by the from_depth/to_depth
            interval are returned.
        """
        # Specify a name if not already passed
        if dataset_name is None:
            dataset_name = '{0}: subdataset {1} to {2}'.format(
                self.name, from_depth, to_depth)

        # Select a data mask
        indices = numpy.where(
            numpy.logical_and(self.from_depths >= from_depth,
                              self.to_depths <= to_depth))

        # Generate a new IntervalDataSet
        newdom = IntervalDataSet(dataset_name,
                                 self.from_depths[indices],
                                 self.to_depths[indices])
        for prop in self.properties.values():
            newdom.add_property(
                property_type=prop.property_type,
                values=prop.values[indices])
        return newdom

    def split_at_gaps(self):
        """ Split a dataset by finding significant gaps in the dataset.

            A metric to define 'gaps' is required. Currently we only have
            'spacing_median', which is really only suitable for PointDataSet
            instances.

            Available metrics:
                'spacing_median': a gap is defined as a spacing between
                    samples which is an order of magnitude above the median
                    sample spacing in a dataset.
        """
        # Generate gap locations
        gap_indices = numpy.flatnonzero(
            self.from_depths[1:] - self.to_depths[:-1])

        # Aggregate gap intervals
        self.gaps = []
        for idx in gap_indices:
            self.gaps.append((self.to_depths[idx], self.from_depths[idx + 1]))

        # We need to include the start and end of the dataset!
        gap_indices = [0] + list(gap_indices) + [-1]

        # Form a list of data subdatasets
        self.subdatasets = []
        for idx in range(len(gap_indices) - 1):
            # DataSets start _after_ the gap & end with next gap
            self.subdatasets.append((
                self.from_depths[gap_indices[idx] + 1],
                self.to_depths[gap_indices[idx + 1]]))
        return self.subdatasets, self.gaps

    def to_point_dataset(self, name=None, depths='midpoint'):
        """ Convert an IntervalDataSet to a PointDataSet

            Uses the specified method to recalculate the depths of the
            PointDataSet points. Defaults to using the midpoint of each
            interval.

            Available depth converters:
                'midpoint': Uses the midpoint of the sampled interval
                'from': Uses the upper depth of the sampled interval
                'to': Uses the lower depths of the sampled interval

            :param name: The identifier for the new dataset. Optional, defaults
                to the same name as the interval dataset.
            :type name: string
            :param depths: the depth converter to use
            :returns: the new PointDataSet instance
        """
        # Generate name
        if name is None:
            name = self.name

        # Generate new depths
        if depths == 'midpoint':
            depths = (self.from_depths + self.to_depths) / 2.
        elif depths == 'from':
            depths = self.from_depths
        elif depths == 'to':
            depths = self.to_depths
        else:
            raise ValueError(
                'Unknown depth conversion method {0}'.format(depths))

        # Generate new dataset
        sdom = PointDataSet(name=name, depths=depths)
        sdom.properties = self.properties.copy()
        return sdom

    def to_dataframe(self):
        """ Tranform the data in the dataset into a Pandas dataframe.
        """
        return pandas.DataFrame(
            data=dict(((k, self.properties[k].values)
                       for k in self.properties.keys())),
            index=zip(self.from_depths, self.to_depths))



""" file:   unmarshallers.py (pysiss.vocabulary.gml)
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: Unmarshalling functions for GeoSciML/GML objects
"""

from ..namespaces import NamespaceRegistry

NAMESPACES = NamespaceRegistry()


UNMARSHALLERS = {}

__all__ = (UNMARSHALLERS,)



""" file: xml_namespaces.py
    author: Jess Robertson
            CSIRO Mineral Resources Flagship
    date:   Monday 25 August, 2014

    description: namespace handling for XML and etree formats

    TODO: Pull unknown namespace definitions from XML file and add to registry
"""

import simplejson
import pkg_resources
from ..utilities import Singleton


class NamespaceRegistry(dict):

    """ Registry for namespace objects
    """

    __metaclass__ = Singleton

    # Load default namespaces list in namespaces.json
    default_namespaces = simplejson.load(
        pkg_resources.resource_stream(
            "pysiss.vocabulary.resources", "namespaces.json"))

    def __init__(self):
        super(NamespaceRegistry, self).__init__()
        self.update(self.default_namespaces)
        self.inverse = dict(reversed(item) for item in self.items())

    def __setitem__(self, key, value):
        super(NamespaceRegistry, self).__setitem__(key, value)
        self.inverse[value] = key

    def __delitem__(self, key):
        del self.inverse[self[key]]
        super(NamespaceRegistry, self).__delitem__(key)


_NAMESPACE_REGISTRY = NamespaceRegistry()


def add_namespace(abbrev, url):
    """ Add an XML namespace to the registry
    """
    _NAMESPACE_REGISTRY[abbrev] = url


def shorten_namespace(tag):
    """ Strip a namespace out of an XML tag and replace it with the shortcut
        version
    """
    ns, tag = split_namespace(tag)
    return _NAMESPACE_REGISTRY.inverse[ns] + ':' + tag


def expand_namespace(tag, form='xml'):
    """ Expand a tag's namespace
    """
    ns, tag = split_namespace(tag)
    if form == 'xml':
        return '{' + _NAMESPACE_REGISTRY[ns] + '}' + tag
    elif form == 'rdf':
        return _NAMESPACE_REGISTRY[ns] + ':' + tag
    else:
        raise ValueError(
            ('Unknown format type {0}. '
             'Allowed values are {1}').format(form, ['xml', 'rdf']))


def split_namespace(tag):
    """ Split a tag into a namespace and a tag
    """
    if tag.startswith('{'):
        # We have an expanded namespace to deal with
        ns, tag = tag.lstrip('{').split('}')
    elif tag.count(':') > 1:
        # We have an expanded namespace of the form ns:ns:ns:ns:tag
        tokens = tag.split(':')
        tag = tokens.pop()
        ns = ':'.join(tokens)
    elif tag.count(':') == 1:
        # We have a shortened namespaces of the form ns:tag
        ns, tag = tag.split(':')
    else:
        # Don't know what to do here? Just return None for namespace
        ns = None
    return ns, tag
